"""
A module for building and performing inference with cluster graphs
"""

# Standard imports
import collections

# Third-party imports
import IPython
import graphviz
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import networkx as nx
import numpy as np
from tqdm.auto import tqdm

# Local imports
from veroku._cg_helpers._cluster import Cluster
import veroku._cg_helpers._animation as cg_animation
from veroku.factors._factor_utils import get_subset_evidence

# TODO: Optimise _pass_message.
# TODO: Improve sepsets selection for less loopiness.
# TODO: Optimisation: messages from clusters that did not receive any new messages in the previous round, do not need
#  new messages calculated.

# pylint: disable=protected-access

DEFAULT_FIG_SIZE = [15, 5]


def _sort_almost_sorted(almost_sorted_deque, key):
    """
    Sort a deque like that where only the first element is potentially unsorted and should probably be last and the rest
     of the deque is sorted in descending order.

    :param collections.deque almost_sorted_deque: The deque of size n, where the first n-1 elements are definitely
        sorted (in descending order) and where the last element is also probably in the correct place, but needs to be
        checked
    :param callable key: The key (function) to use for sorting.
    :return: The sorted (given that the conditions are met) deque.
    :rtype: collections.deque
    """
    if key(almost_sorted_deque[0]) < key(almost_sorted_deque[1]):
        almost_sorted_deque.append(almost_sorted_deque.popleft())
    if key(almost_sorted_deque[-1]) <= key(almost_sorted_deque[-2]):
        return almost_sorted_deque
    almost_sorted_deque = collections.deque(sorted(almost_sorted_deque, key=key, reverse=True))
    return almost_sorted_deque


def _evidence_reduce_factors(factors, evidence):
    """
    Observe relevant evidence for each factor.

    :param factors: The factors to reduce with the (relevant) evidence.
    :type factors: Factor list
    :param dict evidence: The evidence (i.e {'a':1.0, 'b':2.0})
    :return: The reduced factors.
    :rtype factors: Factor list
    """
    reduced_factors = []
    for factor in factors:
        if evidence is not None:
            vrs, values = get_subset_evidence(all_evidence_dict=evidence, subset_vars=factor.var_names)
            if len(vrs) > 0:
                factor = factor.reduce(vrs, values)
        reduced_factors.append(factor.copy())
    return reduced_factors


def _absorb_subset_factors(factors):
    """
    Absorb any factors that has a scope that is a subset of another factor into such a factor.

    :param factors: The list of factors to check for subset factors.
    :type factors: Factor list
    :return: The (potentially reduced) list of factors.
    :rtype: Factor list
    """
    # TODO: Simplify this, if possible.
    factors_absorbtion_dict = {i: [] for i in range(len(factors))}
    final_graph_cluster_factors = []
    # factors: possibly smaller list of factors after factors which have a scope that is a subset of another factor have
    # been absorbed by the larger one.
    factor_processed_mask = [0] * len(factors)
    for i, factor_i in enumerate(factors):
        if not factor_processed_mask[i]:
            factor_product = factor_i.copy()
            for j, factor_j in enumerate(factors):
                if (i != j) and (not factor_processed_mask[j]):
                    if set(factor_j.var_names) < set(factor_product.var_names):
                        factor_product = factor_product.multiply(factor_j)
                        factors_absorbtion_dict[i].append(j)
                        factor_processed_mask[j] = 1
                        factor_processed_mask[i] = 1

            if factor_processed_mask[i]:
                final_graph_cluster_factors.append(factor_product)
    for i, factor_i in enumerate(factors):  # add remaining factors
        if not factor_processed_mask[i]:
            factor_processed_mask[i] = 1
            final_graph_cluster_factors.append(factor_i)
    assert all(
        factor_processed_mask
    ), "Error: Some factors where not included during variable subset processing."
    return final_graph_cluster_factors


class ClusterGraph:
    """
    A class for building and performing inference with cluster graphs.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, factors, evidence=None, special_evidence=None, disable_tqdm=False):
        """
        Construct a Cluster graph from a list of factors.

        :param factors: The factors to construct the graph from
        :type factors: factor list
        :param dict evidence: evidence dictionary (mapping variable names to values) that should be used to reduce
            factors before building the cluster graph.  Example: {'a': 2, 'b':1}
        :param dict special_evidence: evidence dictionary (mapping variable names to values) that should be used in the
            calculation of messages, and not to reduce factors. This allows factor approximations - such as the
            non-linear Gaussian to be iteratively refined. Example: {'a': 2, 'b':1}
        :param bool disable_tqdm: Disable the tqdm progress bars used in graph construction and processing.
        :param bool verbose: Whether or not to output additional information messages during graph construction and
        processing.
        :param debug:
        """
        # TODO: see if evidence and special_evidence can be replaced by a single variable.
        self.num_messages_passed = 0
        self.disable_tqdm = disable_tqdm
        self.verbose = False
        self.debug = False

        self.sync_message_passing_max_distances = []
        if special_evidence is None:
            special_evidence = dict()
        self.special_evidence = special_evidence
        all_evidence_vars = set(self.special_evidence.keys())
        if evidence is not None:
            evidence_vars = set(evidence.keys())
            all_evidence_vars = all_evidence_vars.union(evidence_vars)
        all_factors_copy = _evidence_reduce_factors(factors, evidence)
        final_graph_cluster_factors = _absorb_subset_factors(all_factors_copy)

        clusters = [
            Cluster(factor, cluster_name_prefix=f"c{i}#")
            for i, factor in enumerate(final_graph_cluster_factors)
        ]

        self._set_non_rip_sepsets_dict(clusters, all_evidence_vars)
        self._clusters = clusters

        # Add special evidence to factors
        for cluster in self._clusters:
            cluster_special_evidence_vars, cluster_special_evidence_values = get_subset_evidence(
                self.special_evidence, cluster.var_names
            )
            cluster_special_evidence = dict(
                zip(cluster_special_evidence_vars, cluster_special_evidence_values)
            )
            cluster.add_special_evidence(cluster_special_evidence)

        self.graph_message_paths = collections.deque([])
        self._build_graph()

        # TODO: consolidate these two, if possible
        self.message_passing_animation_frames = []
        self.passed_messages = []

    def _set_non_rip_sepsets_dict(self, clusters, all_evidence_vars):
        """
        Calculate the preliminary sepsets dict before the RIP property is enforced.

        :param clusters: The clusters for which the sepsets should be calculated.
        :param all_evidence_vars: The variables for which there is observed evidence.
        """
        self._non_rip_sepsets = {}
        for i in tqdm(range(len(clusters)), disable=self.disable_tqdm):
            vars_i = clusters[i].var_names
            for j in range(i + 1, len(clusters)):
                vars_j = clusters[j].var_names
                sepset = set(vars_j).intersection(set(vars_i)) - all_evidence_vars
                self._non_rip_sepsets[(i, j)] = sepset
                self._non_rip_sepsets[(j, i)] = sepset

    def _build_graph(self):
        """
        Add the cluster sepsets, graphviz graph and animation graph (for message_passing visualisation).
        """
        # Check for non-unique cluster_ids (This should never be the case)
        cluster_ids = [cluster.cluster_id for cluster in self._clusters]
        if len(set(cluster_ids)) != len(cluster_ids):
            raise ValueError(f"Non-unique cluster ids: {cluster_ids}")

        self._conditional_print("Info: Building graph.")
        self._graph = graphviz.Graph(format="png")
        rip_sepsets_dict = self._get_running_intersection_sepsets()

        # TODO: see why this is necessary, remove if not
        for i in tqdm(range(len(self._clusters)), disable=self.disable_tqdm):
            self._clusters[i].remove_all_neighbours()

        self._conditional_print(f"Debug: number of clusters: {len(self._clusters)}")
        for i in tqdm(range(len(self._clusters)), disable=self.disable_tqdm):

            node_i_name = self._clusters[i]._cluster_id
            self._graph.node(
                name=node_i_name, label=node_i_name, style="filled", fillcolor="white", color="black"
            )
            for j in range(i + 1, len(self._clusters)):

                if (i, j) in rip_sepsets_dict:
                    sepset = rip_sepsets_dict[(i, j)]
                    assert len(sepset) > 0, "Error: empty sepset"
                    self._clusters[i].add_neighbour(self._clusters[j], sepset=sepset)
                    self._clusters[j].add_neighbour(self._clusters[i], sepset=sepset)

                    gmp_ij = _GraphMessagePath(self._clusters[i], self._clusters[j])
                    gmp_ji = _GraphMessagePath(self._clusters[j], self._clusters[i])
                    self.graph_message_paths.append(gmp_ij)
                    self.graph_message_paths.append(gmp_ji)
                    self._clusters[i].add_outward_message_path(gmp_ij)
                    self._clusters[j].add_outward_message_path(gmp_ji)

                    # Graph animation
                    node_j_name = self._clusters[j]._cluster_id
                    sepset_node_label = ",".join(sepset)
                    sepset_node_name = cg_animation.make_sepset_node_name(node_i_name, node_j_name)
                    self._graph.node(name=sepset_node_name, label=sepset_node_label, shape="rectangle")
                    self._graph.edge(node_i_name, sepset_node_name, color="black", penwidth="2.0")
                    self._graph.edge(sepset_node_name, node_j_name, color="black", penwidth="2.0")
        self._conditional_print(f"num graph message paths: {len(self.graph_message_paths)}")

    def _conditional_print(self, message):
        """
        Print message if verbose is True.

        :param message: The message to print.
        """
        if self.verbose:
            print(message)

    def plot_next_messages_info_gain(self, legend_on=False, figsize=None):
        """
        Plot the information gained by a receiving new messages over sebsequent iterations for all message paths in the
            graph.

        :param bool legend_on: Whether or not to show the message paths (specified by connected cluster pairs) in the
            plot legend.
        :param list figsize: The matplotlib figure size.
        """
        if figsize is None:
            figsize = DEFAULT_FIG_SIZE
        plt.figure(figsize=figsize)
        all_paths_information_gains_with_iters = [
            gmp.information_gains_with_iters for gmp in self.graph_message_paths
        ]
        for paths_information_gains_with_iters in all_paths_information_gains_with_iters:
            plt.plot(paths_information_gains_with_iters)
        plt.title("Information Gain of Messages along Graph Message Paths")
        plt.xlabel("iteration")
        plt.ylabel("D_KL(prev_msg||msg)")
        if legend_on:
            legend = [
                f"{gmp.sender_cluster.cluster_id}->{gmp.receiver_cluster.cluster_id}"
                for gmp in self.graph_message_paths
            ]
            plt.legend(legend)

    def plot_message_convergence(self, log=False, figsize=None):
        """
        Plot the the KL-divergence between the messages and their previous instances to indicate the message passing
        convergence.

        :param bool log: If True, plot the log of the KL-divergence.
        :param list figsize: The matplotlib [width, height] of the figure.
        """
        if figsize is None:
            figsize = DEFAULT_FIG_SIZE
        mp_max_dists = self.sync_message_passing_max_distances
        if log:
            mp_max_dists = np.log(mp_max_dists)

        # here we tile an flatten to prevent the plot omission of values with inf on either side.
        mp_max_dists = np.tile(mp_max_dists, [2, 1]).flatten(order="F")
        num_iterations = len(mp_max_dists)

        iterations = np.array(list(range(num_iterations))) / 2  # divide by 2 to correct for tile and flatten
        non_inf_max_distances = [d for d in mp_max_dists if d != np.inf]
        max_non_inf = max(non_inf_max_distances)
        new_inf_value = max_non_inf * 1.5
        max_distances_replaces_infs = np.array([v if v != np.inf else new_inf_value for v in mp_max_dists])
        inf_values = np.ma.masked_where(
            max_distances_replaces_infs != new_inf_value, max_distances_replaces_infs
        )
        plt.figure(figsize=figsize)
        plt.plot(iterations, max_distances_replaces_infs)
        plt.plot(iterations, inf_values, c="r", linewidth=2)
        if len(non_inf_max_distances) != len(mp_max_dists):
            custom_lines = [Line2D([0], [0], color="r", lw=4)]
            plt.legend(custom_lines, ["infinity"])
        plt.title("Message Passing Convergence")
        plt.xlabel("iteration")
        plt.ylabel("log max D_KL(prev_msg||msg)")
        plt.show()

    def _get_unique_vars(self):
        """
        Get the set of variables in the graph.

        :return: The variables
        :rtype: list
        """
        all_vars = []
        for cluster in self._clusters:
            all_vars += cluster.var_names
        unique_vars = list(set(all_vars))
        return unique_vars

    def _get_vars_min_spanning_trees(self):
        """
        Get the minimum spanning trees of all the variables in the graph.
        """
        all_vars = self._get_unique_vars()
        var_graphs = {var: nx.Graph() for var in all_vars}
        num_clusters = len(self._clusters)
        for i in range(num_clusters):
            for j in range(i + 1, num_clusters):
                sepset = self._non_rip_sepsets[(i, j)]
                for var in sepset:
                    var_graphs[var].add_edge(i, j, weight=1)
        var_spanning_trees = dict()
        for var in all_vars:
            var_graph = var_graphs[var]
            var_spanning_trees[var] = nx.minimum_spanning_tree(var_graph)
        return var_spanning_trees

    def _get_running_intersection_sepsets(self):
        """
        Get a set of sepsets for the graph, such that the graph, with these sepsets satisfies the running intersection
        property.
        """
        edge_sepset_dict = {}
        unique_vars = self._get_unique_vars()
        min_span_trees = self._get_vars_min_spanning_trees()
        self._conditional_print("Info: Getting unique variable spanning trees.")
        for i in tqdm(range(len(unique_vars)), disable=self.disable_tqdm):
            var = unique_vars[i]
            min_span_tree = min_span_trees[var]
            for edge in min_span_tree.edges():
                if edge in edge_sepset_dict:
                    edge_sepset_dict[edge].append(var)
                else:
                    edge_sepset_dict[edge] = [var]
        return edge_sepset_dict

    def show(self):
        """
        Show the cluster graph.
        """
        self._graph.render("/tmp/test.gv", view=False)
        image = IPython.core.display.Image("/tmp/test.gv.png")
        IPython.core.display.display(image)

    def save_graph_image(self, filename):
        """
        Save image of the graph.

        :param filename: The filename of the file.
        """
        graphviz.Source(self._graph, filename=filename, format="png")

    def get_marginal(self, vrs):
        """
        Search the graph for a specific variable and get that variables marginal (posterior marginal if process_graph
        has been run previously).

        :return: The marginal
        :rtype: Factor child
        """
        for cluster in self._clusters:
            if set(vrs) <= set(cluster.var_names):
                factor = cluster.factor
                evidence_vrs, evidence_values = get_subset_evidence(self.special_evidence, factor.var_names)
                if len(evidence_vrs) > 0:
                    factor = factor.reduce(evidence_vrs, evidence_values)
                marginal = factor.marginalize(vrs, keep=True)
                return marginal
        raise ValueError(f"No cluster with variables containing {vrs}")

    def get_posterior_joint(self):
        """
        Get the posterior joint distribution. This function is only intended to be used as a research / debugging tool
        for small networks.
        """
        # TODO: add functionality for efficiently getting a posterior marginal over any subset of variables and replace
        #  the get_marginal function above.
        cluster_product = self._clusters[0]._factor.joint_distribution
        for cluster in self._clusters[1:]:
            cluster_product = cluster_product.multiply(cluster._factor.joint_distribution)
        last_passed_message_factors = self._last_passed_message_factors
        if len(last_passed_message_factors) == 0:
            assert self.num_messages_passed == 0
            joint = cluster_product
        else:
            message_product = last_passed_message_factors[0]
            for message_factor in last_passed_message_factors[1:]:
                message_product = message_product.multiply(message_factor)
            joint = cluster_product.cancel(message_product)
        return joint

    def process_graph(self, tol=1e-3, max_iter=50, make_animation_gif=False):
        """
        Perform synchronous message passing until convergence (or maximum iterations).

        :param tol: The minimum tolerance value for the KL divergence D_KL(previous_message || next_message) that needs
            to be reached (for all messages) before stopping message passing (before max_iter is reached).
        :param max_iter: The maximum number of iterations of message passing. The maximum number of messages that can be
            passed is max_iter * n, where n is the number of message paths (2x the number of edges) in the graph.
        :param bool make_animation_gif: Whether or not to create an animation of the message passing process when. Note:
            This can cause slow processing and high memory consumption for large graphs and is therefore recommended to
            be used only with very small (>50 cluster) graphs.
        """

        self.sync_message_passing_max_distances = []
        if len(self._clusters) == 1:
            # The Cluster Graph contains only single cluster. Message passing not possible or necessary.
            if self.special_evidence:
                evidence_vrs = list(self.special_evidence.keys())
                evidence_values = list(self.special_evidence.values())
                self._clusters[0]._factor = self._clusters[0]._factor.reduce(
                    vrs=evidence_vrs, values=evidence_values
                )
            return

        # TODO: see if the definition of max_iter can be improved
        key_func = lambda x: x.next_information_gain
        max_message_passes = max_iter * len(self.graph_message_paths)

        self.graph_message_paths = collections.deque(
            sorted(self.graph_message_paths, key=key_func, reverse=True)
        )

        for _ in tqdm(range(max_message_passes), disable=self.disable_tqdm):
            sender_cluster_id = self.graph_message_paths[0].sender_cluster.cluster_id
            receiver_cluster_id = self.graph_message_paths[0].receiver_cluster.cluster_id
            if self.debug:
                self.passed_messages.append(self.graph_message_paths[0].next_message.copy())
            self.graph_message_paths[0].pass_next_message()
            self.num_messages_passed += 1
            self.graph_message_paths = collections.deque(
                sorted(self.graph_message_paths, key=key_func, reverse=True)
            )
            # TODO: test this:
            # self.graph_message_paths = _sort_almost_sorted(self.graph_message_paths, key=key_func)

            max_next_information_gain = self.graph_message_paths[0].next_information_gain
            self.sync_message_passing_max_distances.append(max_next_information_gain)
            if max_next_information_gain <= tol:
                return

            if make_animation_gif:
                cg_animation.add_message_pass_animation_frames(
                    graph=self._graph,
                    frames=self.message_passing_animation_frames,
                    node_a_name=sender_cluster_id,
                    node_b_name=receiver_cluster_id,
                )

    @property
    def _last_passed_message_factors(self):
        """
        The factors of the messages passed in the last iteration of message passing.
        """
        return [gmp.previously_sent_message.factor for gmp in self.graph_message_paths]

    def make_message_passing_animation_gif(self, filename="graph_animation.gif"):
        """
        Make message passing animation and save a GIF of the animation to file. Note that this function will only work
        if the make_animation_gif variable was set to True when the process_graph method was called.

        :param str filename: The name of the file.
        """
        self.message_passing_animation_frames[0].save(
            fp=f"./{filename}",
            format="GIF",
            append_images=self.message_passing_animation_frames[1:],
            save_all=True,
            duration=400,
            loop=0,
        )


class _GraphMessagePath:
    """
    A specific path (direction along an edge) in a graph along which a message can be passed.
    """

    def __init__(self, sender_cluster, receiver_cluster):
        """
        The initializer.

        :param Cluster sender_cluster: The cluster that defines the starting point of the path.
        :param Cluster  receiver_cluster: The cluster that defines the end point of the path.
        """
        self.sender_cluster = sender_cluster
        self.receiver_cluster = receiver_cluster
        self.previously_sent_message = None
        self.next_message = self.sender_cluster.make_message(self.receiver_cluster.cluster_id)
        self.next_information_gain = None
        self.information_gains_with_iters = []
        self.update_next_information_gain()

    def update_next_information_gain(self):
        """
        Calculate the information gain that will be achieved when passing the next message.
        """
        if self.previously_sent_message is None:
            self.next_information_gain = self.next_message.distance_from_vacuous()
        else:
            # "In the context of machine learning, KL(P||Q) is often called the information gain achieved if Q is
            # used instead of P." - wikipedia
            # We typically want to know which new message (Q) will result in the largest information gain if it replaces
            # the message (P)
            # message: previous_message (P)
            # factor: next message (Q)
            # P.kl_divergence(Q)
            self.next_information_gain = self.previously_sent_message.kl_divergence(self.next_message)
        self.information_gains_with_iters.append(self.next_information_gain)

    def recompute_next_message(self):
        """
        Recompute the next message.
        """
        new_next_message = self.sender_cluster.make_message(self.receiver_cluster.cluster_id)
        self.next_message = new_next_message.copy()
        self.update_next_information_gain()

    def pass_next_message(self):
        """
        Pass the next message along this path.
        """
        self.receiver_cluster.receive_message(self.next_message)
        self.previously_sent_message = self.next_message.copy()
        self.next_information_gain = 0.0
        self.information_gains_with_iters.append(self.next_information_gain)

        for gmp in self.receiver_cluster._outward_message_paths:
            gmp.recompute_next_message()



"""
A module for computing and transforming sigma points and converting between sigma points and covariance parameters.
"""

import numpy as np
import veroku.factors.gaussian as gauss


def get_sigma_points_array(gaussian_factor):
    """
    Get 2*dim + 1 equals weight sigma points for a Gaussian distribution with the given parameters.

    :param gaussian_factor: The factor
    :return: (numpy array) An array with sigma points as column vectors.
    """
    assert isinstance(gaussian_factor, gauss.Gaussian), "Error: gaussian_factor is not a Gaussian factor."
    cov = gaussian_factor.get_cov()
    mean = gaussian_factor.get_mean()
    chol = np.linalg.cholesky(cov)
    dim = cov.shape[0]
    k = (dim + 0.5) ** 0.5
    sigma_point_array_n = mean - k * chol
    sigma_point_array_p = mean + k * chol
    sigma_points_array = np.concatenate([sigma_point_array_n, mean, sigma_point_array_p], axis=1)
    return sigma_points_array


def sigma_point_array_to_gaussian_params(sigma_point_array):
    """
    Convert a sigma point array to Gaussian covariance parameters.

    :param sigma_point_array: (numpy array) An array where the columns correspond to sigma points.
    :return: The mean (numpy array) and covariance (numpy array) of the sigma points.
    """
    mean = np.expand_dims(np.mean(sigma_point_array, axis=1), axis=1)
    cov = np.cov(sigma_point_array, bias=1)
    return cov, mean


def sigma_points_to_gaussian_params(sigma_points):
    """
    Convert a sigma points to Gaussian covariance parameters.

    :param sigma_points: (list of numpy arrays) An array where the columns correspond to sigma points.
    :return: The mean (numpy array) and covariance (numpy array) of the sigma points.
    """
    sigma_point_array = np.concatenate(sigma_points, axis=1)
    return sigma_point_array_to_gaussian_params(sigma_point_array)


def sigma_points_array_to_joint_params(sigma_points_array, transform, var_names=None):
    """
    Construct joint sigma points using the original sigma points and the transform function, then Calculate the joint
    covariance parameters of the joint sigma points.

    :param sigma_points_array: (numpy array) An array where the columns correspond to sigma points.
    :param transform: (function) The transformation function.
    :param var_names: The variable names of the variables to be transformed.
    :return: The joint mean (numpy array) and joint covariance (numpy array) of the sigma points.
    """
    transformed_sigma_points_array = transform(sigma_points_array, var_names)
    joint_sigma_point_array = np.concatenate([sigma_points_array, transformed_sigma_points_array], axis=0)
    joint_cov, joint_mean = sigma_point_array_to_gaussian_params(joint_sigma_point_array)
    return joint_cov, joint_mean



"""
A module for animating graphviz graphs.
"""
# Standard imports
import re

# Third-party imports
from PIL import Image as PIL_Image

# pylint: disable=anomalous-backslash-in-string


def make_sepset_node_name(node_a_name, node_b_name):
    """
    Make a standard sepset node name using the two neighboring nodes.

    :param str node_a_name: The one node's name.
    :param str node_b_name: The other node's name.
    :return: The sepset's name
    :rtype: str
    """
    return "sepset__" + "__".join(sorted([node_a_name, node_b_name]))


def change_cluster_graph_edge_color(graph, node_a_name, node_b_name, new_color):
    """
    Change the color of an edge in a cluster graph graphviz object. Note that the cluster graph has cluster and sepset
    nodes. The 'edge' of which the color is changed in this function is actually two edges in the graphviz object (the
    edge between node_a and the sepset node and the edge between the sepset node and node_b).

    :param graph: The graphviz graph
    :param node_a_name: The edge's one neighboring node's name.
    :param node_b_name: The edge's other neighboring node's name.
    :param new_color: The new color (i.e 'green' or 'red')
    """

    sepset_node_name = make_sepset_node_name(node_a_name, node_b_name)
    change_graph_edge_color(
        graph=graph, node_a_name=node_a_name, node_b_name=sepset_node_name, new_color=new_color
    )
    change_graph_edge_color(
        graph=graph, node_a_name=node_b_name, node_b_name=sepset_node_name, new_color=new_color
    )


def sub_and_super_both_in(sub_string, super_string, main_string):
    """
    Check if a substring and a super (contains the substring) string are both in another string in non overlapping areas.

    :param sub_string: The string contained in the super-string.
    :param super_string: The string containing the sub-string.
    :param main_string: The main string to check the independent presence of the strings.
    :return:
    """
    if sub_string not in main_string:
        return False
    reduced_string = main_string.replace(super_string, "")
    if sub_string in reduced_string:
        return True
    return False


def contains_non_overlapping_substrings(substring_a, substring_b, main_string):
    """
    Check that both sub-strings can be found in separate locations in strings

    :param substring_a: The one substring to check for.
    :param substring_b: The other substring to check for.
    :param main_string: The string to check in.
    :return: Result of check for independent string presence.

    :Examples:

    >>> substring_a = "abc"
    >>> substring_b = "123"
    >>> main_string = "abc123 -- def"
    >>> contains_non_overlapping_substrings(substring_a, substring_b, string)
    >>> False

    >>> substring_a = "abc"
    >>> substring_b = "123"
    >>> main_string = "abc -- 123"
    >>> contains_non_overlapping_substrings(substring_a, substring_b, string)
    >>> True
    """

    if substring_a in substring_b:
        return sub_and_super_both_in(substring_a, substring_b, main_string)

    if substring_b in substring_a:
        return sub_and_super_both_in(substring_b, substring_a, main_string)
    # Neither substring strings is a substring of the other, so we can remove the one and check if the other is still
    # present.
    if substring_a in main_string:
        reduced_string = main_string.replace(substring_a, "")
    else:
        return False
    result = substring_b in reduced_string
    return result


def change_graph_edge_color(graph, node_a_name, node_b_name, new_color):
    """
    Change the edge color of the edge between two nodes in a graphviz object.

    :param graph: The graphviz graph
    :param node_a_name: The one node
    :param node_b_name: The other node
    :param new_color: The new color of the edge (i.e 'green', 'blue', 'red')
    """

    edge_found = False
    for i, string_i in enumerate(graph.body):
        if "--" in string_i:
            if contains_non_overlapping_substrings(node_a_name, node_b_name, string_i):
                pattern = f'(\\t?"?{node_a_name}"?\s--\s"?{node_b_name}"?\s\[.*color=)([\S]*\s)(.*)'
                new_string_i = re.sub(pattern, f"\g<1>{new_color} \g<3>", string_i)
                # TODO: improve this (it's a bit hacky)
                if new_string_i == string_i:
                    pattern = f'(\\t?"?{node_b_name}"?\s--\s"?{node_a_name}"?\s\[.*color=)([\S]*\s)(.*)'
                    new_string_i = re.sub(pattern, f"\g<1>{new_color} \g<3>", string_i)

                graph.body[i] = new_string_i
                edge_found = True
    if not edge_found:
        raise ValueError(f"no edge between node {node_a_name} and node {node_b_name} in graph.")


def change_graph_node_color(graph, node_name, new_color):
    """
    Change the fill color of a node in a cluster graph graphviz object.

    :param graph:  The graphviz graph
    :param node_name: The name of the node which colour should be changed.
    :param new_color: The new fill color of the node (i.e 'green', 'blue', 'red')
    :return:
    """
    node_found = False
    for i, element_string in enumerate(graph.body):
        if "--" not in element_string:  # not edge
            if node_name in element_string:
                graph.body[i] = re.sub(
                    f'(\\t?"?{node_name}"?\s\[.*fillcolor=)([\S]*\s)(.*)', f"\g<1>{new_color} \g<3>", element_string
                )
                node_found = True
    if not node_found:
        raise ValueError(f"cannot change colour of not existing node: no node {node_name} in graph")


def graph_to_pil_image(graph):
    """
    Convert the graph to a PIL image.
    :param graph: The graph to convert.
    """
    graph.render("/tmp/test.gv", view=False)
    pil_image = PIL_Image.open("/tmp/test.gv.png")
    pil_image_copy = pil_image.copy()
    pil_image.close()
    return pil_image_copy


def add_message_pass_animation_frames(graph, frames, node_a_name, node_b_name):
    """
    Add frames (as part of the animation) representing a message being passed between two nodes in the graph.
    :param graph: The graph.
    :param frames: The frames to add.
    :param node_a_name: The one node.
    :param node_b_name: The other node.
    """
    new_color = "red"
    change_graph_node_color(graph=graph, node_name=node_a_name, new_color=new_color)
    frames.append(graph_to_pil_image(graph))

    change_cluster_graph_edge_color(
        graph=graph, node_a_name=node_a_name, node_b_name=node_b_name, new_color=new_color
    )
    frames.append(graph_to_pil_image(graph))

    change_graph_node_color(graph=graph, node_name=node_b_name, new_color=new_color)
    frames.append(graph_to_pil_image(graph))
    change_graph_node_color(graph=graph, node_name=node_a_name, new_color="white")
    frames.append(graph_to_pil_image(graph))

    change_cluster_graph_edge_color(
        graph=graph, node_a_name=node_a_name, node_b_name=node_b_name, new_color="black"
    )
    frames.append(graph_to_pil_image(graph))
    change_graph_node_color(graph=graph, node_name=node_b_name, new_color="white")
    frames.append(graph_to_pil_image(graph))



"""
This module contains an abstract parent class, defining the minimum functions that all factors should have
"""

from abc import ABCMeta, abstractmethod
import copy

from veroku._constants import DEFAULT_FACTOR_RTOL, DEFAULT_FACTOR_ATOL


class Factor:
    """
    An abstract parent class.
    """

    __metaclass__ = ABCMeta

    def __init__(self, var_names):
        """
        A super class constructor that should be called from the base class constructor.
        """
        if len(set(var_names)) != len(var_names):
            raise ValueError("duplicate variables in var_names: ", var_names)

        self._var_names = var_names
        if not isinstance(var_names, list):
            self._var_names = [var_names]

        self._dim = len(var_names)

    @property
    def var_names(self):
        """
        Get the factor's variable names.

        :return: the var_names parameter.
        """
        return copy.deepcopy(self._var_names)

    @property
    def dim(self):
        """
        Get the factor's dimensionality.

        :return: the dim parameter.
        """
        return self._dim

    def get_marginal_vars(self, vrs, keep=True):
        """
        A helper function (for marginalize) that returns the variables that should be marginalised out (keep=False)
        or kept (keep=True).

        :param vrs: (list) the variables
        :param keep: Whether these variables are to be kept or summed out.
        :return: The variables to be kept or summed out.
        """
        if isinstance(vrs, str):
            vrs = [vrs]
        if keep:
            return vrs.copy()
        return list(set(self.var_names) - set(vrs))

    @abstractmethod
    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between the message factor and a vacuous (uniform) version of it.

        :return: The KL-divergence
        """

    @property
    def is_vacuous(self):
        """
        Check if this factor is vacuous (i.e uniform).

        :return: Whether the factor is vacuous or not.
        :rtype: bool
        """
        if self.distance_from_vacuous() < 1e-10:
            return True
        return False

    @abstractmethod
    def equals(self, factor, rtol=DEFAULT_FACTOR_RTOL, atol=DEFAULT_FACTOR_ATOL):
        """
         An abstract function for checking if this factor is equal to another factor.

        :param factor: The factor to be compared to.
        :param float rtol: The relative tolerance to use for factor equality check.
        :param float atol: The absolute tolerance to use for factor equality check.
        :return: The result of the check.
        :rtype: bool
        """

    @abstractmethod
    def copy(self):
        """
        An abstract function for copying a factor that should be implemented in the base class.

        :return: a copy of the factor
        """

    @abstractmethod
    def marginalize(self, vrs, keep=True):
        """
        An abstract function for performing factor marginalisation that should be implemented in the base class.

        :return: the resulting marginal factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: The resulting factor marginal.
        """

    @abstractmethod
    def multiply(self, factor):
        """
        An abstract function for performing factor multiplication that should be implemented in the base class.

        :param factor: The factor to be multiplied with.
        :return: The resulting product
        """

    def absorb(self, factor):
        """
        (Alias for multiply) An abstract function for performing factor multiplication that should be implemented in the base class.

        :param factor: The factor to be multiplied with.
        :return: The resulting product
        """

        return self.multiply(factor)

    @abstractmethod
    def divide(self, factor):
        """
        An abstract function for performing factor division that should be implemented in the base class.

        :param factor: The factor to be divided by.
        :return: The resulting quotient
        """

    def cancel(self, factor):
        """
        (Alias for divide by default - but can differ in certain cases) An abstract function for performing factor
        division (or division-like operarations - see Categorical cancel for example) that can be implemented in the
        base class.

        :param factor: The factor to be divided by.
        :return: The resulting factor
        """
        return self.divide(factor)

    @abstractmethod
    def reduce(self, vrs, values):
        """
        An abstract function for performing the observation (a.k.a conditioning) operation that should be implemented
        in the base class.

        :return: the resulting reduced factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param values: The values of vars.
        :return: The resulting reduced factor.
        """

    def observe(self, vrs, values):
        """
        (Alias for reduce) Reduce a factor by observing certain values for certain variables.

        :param vrs: The observed variables.
        :param values: The values of these variables.
        :return: The reduced factor.
        """
        return self.reduce(vrs, values)

    @abstractmethod
    def normalize(self):
        """
        An abstract function for performing factor normalization that should be implemented in the base class.

        :return: The normalized factor.
        """

    @abstractmethod
    def show(self):
        """
        An abstract function for printing the parameters of the factor that should be implemented in the base class.
        """

    @property
    def joint_distribution(self):
        """
        The joint distribution.
        """
        return self.copy()

    def __mul__(self, other):
        """
        Multiply this factor with another factor.

        :param other: The other factor.
        :return: The result.
        """
        return self.multiply(other)

    def __div__(self, other):
        """
        Divide this factor with another factor.

        :param other: The other factor.
        :return: The result.
        """
        return self.divide(other)

    def __eq__(self, other):
        """
        Compare this factor to another factor.

        :param other: The other factor.
        :return: The equality result.
        """
        return self.equals(other)

    def __ne__(self, other):
        """
        Compare this factor to another factor.

        :param other: The other factor.
        :return: The inverse equality result.
        """
        return not self.__eq__(other)



"""
A module for instantiating and performing operations on multivariate Gaussian distributions.
"""

# pylint: disable=cyclic-import
# pylint: disable=protected-access
# pylint: disable=no-self-use

# System imports
import cmath
import copy
import operator

# Third-party imports
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

# Local imports
from veroku.factors._factor import Factor
from veroku.factors import _factor_utils
from veroku.factors._factor_template import FactorTemplate
from veroku._constants import DEFAULT_FACTOR_RTOL, DEFAULT_FACTOR_ATOL
import veroku.factors.experimental.gaussian_mixture as gm


def make_random_gaussian(var_names, mean_range=None, cov_range=None):
    """
    Make a d dimensional random Gaussian by independently sampling the mean and covariance parameters from uniform
    distributions.

    :param var_names: The variable name of the factor
    :type var_names: str list
    :param mean_range: The range between which a mean will the uniformly sampled.
    :type mean_range: float list
    :param cov_range:The range between which a covariance will the uniformly sampled.
    :type cov_range: float list
    :return: The random Gaussian.
    :rtype: Gaussian
    """
    if mean_range is None:
        mean_range = [-10, 10]
    if cov_range is None:
        cov_range = [1, 10]
    assert var_names, "Error: var_names list cannot be empty."
    dim = len(var_names)
    # TODO: Add non diagonal covariance matrix sampling and add test for PSD covariance matrices.
    cov = np.random.uniform(cov_range[0], cov_range[1], [dim, dim]) * np.eye(dim, dim)
    mean = np.random.uniform(mean_range[0], mean_range[1], [dim, 1])
    random_gaussian = Gaussian(cov=cov, mean=mean, log_weight=0.0, var_names=var_names)
    return random_gaussian


def make_std_gaussian(var_names):
    """
    Make a d dimensional standard Gaussian.

    :param var_names: The variable name of the factor.
    :type var_names: str list
    :return: The standard Gaussian
    :rtype: Gaussian
    """
    assert var_names, "Error: var_names list cannot be empty."
    dim = len(var_names)
    cov = np.eye(dim, dim)
    mean = np.zeros([dim, 1])
    random_gaussian = Gaussian(cov=cov, mean=mean, log_weight=0.0, var_names=var_names)
    return random_gaussian


def make_linear_gaussian(a_mat, n_mat, conditioning_vars, conditional_vars):
    """
    Make a linear Gaussian factor.

    :param a_mat: The linear transform matrix.
    :type a_mat: numpy.ndarray
    :param n_mat: The additive noise covariance matrix.
    :type n_mat: numpy.ndarray
    :param conditioning_vars: The list of conditioning variables.
    :param conditional_vars: The list of conditional variables.
    :return: The linear Gaussian Template
    :rtype: Gaussian
    """

    # TODO: improve this by adding the correct equations for calculating the linear Gaussian parameters that do not rely
    #  on the dummy factor multiplication.
    x_dim = len(conditioning_vars)
    cov_xx = np.eye(x_dim)
    prec_xx = np.linalg.inv(cov_xx)

    u_x = np.zeros([x_dim, 1])
    h_x = prec_xx @ u_x
    g_x = -0.5 * u_x.T @ h_x + np.log(1.0) - 0.5 * np.log(np.linalg.det(2.0 * np.pi * cov_xx))

    cov = np.block([[cov_xx, cov_xx @ a_mat.T], [a_mat @ cov_xx, a_mat @ cov_xx @ a_mat.T + n_mat]])
    prec_joint = np.linalg.inv(cov)
    prec_cpd = prec_joint.copy()
    prec_cpd[:x_dim, :x_dim] = prec_cpd[:x_dim, :x_dim] - prec_xx

    mean_joint = np.block([[u_x], [a_mat.T @ u_x]])

    h_cpd = prec_joint @ mean_joint
    h_cpd[:x_dim] = h_cpd[:x_dim] - h_x

    utkux = mean_joint.T @ prec_joint @ mean_joint
    g_joint = 0.5 * utkux + np.log(1.0) - 0.5 * np.log(np.linalg.det(2.0 * np.pi * cov))
    g_cpd = g_joint - g_x

    return Gaussian(var_names=conditioning_vars + conditional_vars, prec=prec_cpd, h_vec=h_cpd, g_val=g_cpd)


def make_linear_gaussian_cpd_template(a_mat, n_mat, conditioning_var_templates, conditional_var_templates):
    """
    Make a linear Gaussian factor template.

    :param a_mat: The linear transform matrix.
    :type a_mat: numpy.ndarray
    :param n_mat: The additive noise covariance matrix.
    :type n_mat: numpy.ndarray
    :param conditioning_var_templates: The list of formattable strings for the conditioning variables (i.e: ['var_a{i}_{t}', 'var_b{i}_{t}'])
    :param conditional_var_templates: The list of formattable strings for the conditional variables (i.e: ['var_c{i}_{t}', 'var_d{i}_{t}'])
    :return: The linear Gaussian Template
    :rtype: Gaussian
    """
    assert len(conditioning_var_templates) == a_mat.shape[0]
    assert n_mat.shape[0] == a_mat.shape[0]
    nlg = make_linear_gaussian(a_mat, n_mat, conditioning_var_templates, conditional_var_templates)
    var_templates = conditioning_var_templates + conditional_var_templates
    gaussian_parameters = {"prec": nlg.get_prec(), "h": nlg.get_h(), "g": 0}
    return GaussianTemplate(gaussian_parameters, var_templates)


class Gaussian(Factor):
    """
    A class for instantiating and performing operations om multivariate Gaussian distributions.
    """

    def __init__(self, cov=None, mean=None, log_weight=None, prec=None, h_vec=None, g_val=None, var_names=None):
        """
        General constructor that can use either covariance form or canonical form parameters to construct a
        d-dimensional multivariate Gaussian object.

        :param cov: (covariance form) The covariance matrix (or variance scalar in 1-dimensional case).
        :type cov: dxd numpy.ndarray or float
        :param mean: (covariance form) The mean vector (or mean scalar in 1-dimensional case).
        :type mean: dx1 numpy.ndarray or float
        :param log_weight: (covariance form) The log of the weight (the Gaussian function integrates to the weight value)
        :type log_weight: float
        :param prec: (canonical form) The precision matrix (or precision scalar in 1-dimensional case).
        :type prec: dxd numpy.ndarray or float
        :param h_vec: (canonical form) The h vector  (or h scalar in 1-dimensional case).
        :type h_vec: dx1 numpy.ndarray or float
        :param g_val: (canonical form) The g parameter.
        :type g_val: float
        :param var_names: a list of variable names where the order of the names correspond to the order in the
                         mean and covariance (or precision and h vector)  parameters.
        :type var_names: str list

        Example:

        .. code-block:: python

            # Using covariance form parameters
            >>> Gaussian(cov=[[1, 0], [0, 1]], mean=[0, 0], log_weight=0.0, var_names=["a", "b"])
            # Using canonical form parameters
            >>> Gaussian(prec=[[1, 0], [0, 1]], h_vec=[0, 0], g_val=0.0, var_names=["a", "b"])

        """
        if var_names is None:
            raise ValueError("var_names not provided.")
        super().__init__(var_names=var_names)
        self._is_vacuous = False
        if all(v is None for v in [prec, h_vec, g_val]):
            if any(v is None for v in [cov, mean, log_weight]):
                raise ValueError("incomplete parameters")
            self.cov = _factor_utils.make_square_matrix(cov)
            self.mean = _factor_utils.make_column_vector(mean)
            assert len(self.cov.shape) == 2, "Error: Covariance matrix must be two dimensional."

            assert self.cov.shape[0] == self.dim, "Error: Covariance matrix dimension inconsistency."
            assert self.mean.shape[0] == self.dim, "Error: Mean vector dimension inconsistency."
            self.log_weight = _factor_utils.make_scalar(log_weight)
            self.prec, self.h_vec, self.g_val = None, None, None
            self.canform = False
            self.covform = True
        else:
            if any(v is None for v in [prec, h_vec, g_val]):
                raise ValueError("incomplete parameters")
            self.prec = _factor_utils.make_square_matrix(prec)
            self.h_vec = _factor_utils.make_column_vector(h_vec)
            if self.prec.shape[0] != self.dim:
                pass
            assert len(self.prec.shape) == 2, "Error: Precision matrix must be two dimensional."
            assert self.prec.shape[0] == self.dim, "Error: Precision matrix dimension inconsistency."
            h_error_msg = f"Error: h vector dimension inconsistency: {self.h_vec.shape[0]} (should be {self.dim})"
            assert self.h_vec.shape[0] == self.dim, h_error_msg
            self.g_val = _factor_utils.make_scalar(g_val)
            self.cov, self.mean, self.log_weight = None, None, None
            self.canform = True
            self.covform = False

            # Note: this is important to allow vacuous Gaussians (that arise, for exmaple, when Gaussians are divided by
            # identical distributions or with unconditioned nonlinear Gaussians) to be merginalised.
            if self.is_vacuous:
                self._is_vacuous = True

    @classmethod
    def make_vacuous(cls, var_names, g_val=0.0):
        """
        Make an vacuous Gaussian distribution with a zero precision matrix and zero h vector and zero g value.
        This 'Gaussian' is effectively a constant function (it has infinite variance and infinite mass) with value
        exp(g).

        :param var_names: The variable names.
        :type var_names: str list
        :param g_val: The g parameter.
        :type g_val: float
        :return: The vacuous Gaussian with zero K and h parameters.
        :rtype: Gaussian
        """
        dim = len(var_names)
        prec_zero = np.zeros([dim, dim])
        h_zero = np.zeros([dim, 1])
        g_zero = g_val
        return cls(prec=prec_zero, h_vec=h_zero, g_val=g_zero, var_names=var_names)

    def _reorder_parameters(self, new_order_vars):
        """
        Reorder the values in the matrix and vector parameters according to the order of new_order_vars

        :param new_order_vars: The same variable names as in self.var_names, in a different order
        :type new_order_vars: str list
        """
        if new_order_vars == self._var_names:
            return
        assert set(new_order_vars) == set(
            self._var_names
        ), "Error: new_order_vars must contain the same variable names as in var_names."
        new_order = [new_order_vars.index(var) for var in self._var_names]

        if self.canform:
            new_prec = np.zeros([self.dim, self.dim])
            new_h_vec = np.zeros([self.dim, 1])
            for i, new_i in enumerate(new_order):
                new_h_vec[new_i, 0] = self.h_vec[i, 0]
                for j, new_j in enumerate(new_order):
                    new_prec[new_i, new_j] = self.prec[i, j]
            self.prec = new_prec
            self.h_vec = new_h_vec

        if self.covform:
            new_cov = np.zeros([self.dim, self.dim])
            new_mean = np.zeros([self.dim, 1])
            for i, new_i in enumerate(new_order):
                new_mean[new_i, 0] = self.mean[i, 0]
                for j, new_j in enumerate(new_order):
                    new_cov[new_i, new_j] = self.cov[i, j]
            self.cov = new_cov
            self.mean = new_mean
        self._var_names = new_order_vars

    def _canform_equals(self, gaussian, rtol, atol):
        """
        Helper function for check equivalence of canonical form parameters.
        """
        if not np.isclose(self.g_val, gaussian.get_g(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        gaussian._reorder_parameters(self.var_names)
        if not np.allclose(self.h_vec, gaussian.get_h(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        if not np.allclose(self.prec, gaussian.get_prec(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        return True

    def _covform_equals(self, gaussian, rtol, atol):
        """
        Helper function for check equivalence of covariance form parameters.
        """

        if not np.isclose(self.get_weight(), gaussian.get_weight(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        gaussian._reorder_parameters(self.var_names)
        if not np.allclose(self.mean, gaussian.get_mean(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        if not np.allclose(self.cov, gaussian.get_cov(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        return True

    def equals(self, factor, rtol=DEFAULT_FACTOR_RTOL, atol=DEFAULT_FACTOR_ATOL):
        """
        Check if this factor is the same as another factor.

        :param factor: The factor to compare with.
        :type factor: Gaussian
        :param float rtol: The absolute tolerance parameter (see numpy Notes for allclose function).
        :param float atol: The absolute tolerance parameter (see numpy Notes for allclose function).
        :return: Result of equals comparison between self and gaussian
        rtype: bool
        """

        # TODO: extend this to cover other factors that could be equal (i.e Gaussian mixtures with one component or
        #  Gaussian factorised factors)
        if not isinstance(factor, Gaussian):
            raise ValueError(f"unexpected factor type {type(factor)} in Gaussian equals function.")

        if set(self._var_names) != set(factor.var_names):
            return False

        gaussian_copy = factor.copy()
        if self._var_names != factor.var_names:
            gaussian_copy._reorder_parameters(self._var_names)
        if gaussian_copy._is_vacuous and self._is_vacuous:
            return True
        if self.canform:
            if self._canform_equals(factor, rtol, atol):
                return True
            return False
        self._update_covform()
        if self.covform:
            if not self._covform_equals(factor, rtol, atol):
                return False
        return True

    def get_prec(self):
        """
        Get the precision parameter.

        :return: The prec parameter.
        :rtype: dxd numpy.ndarray or float
        """
        self._update_canform()
        if self.prec is not None:
            return self.prec.copy()
        return None

    def get_h(self):
        """
        Get the h vector.

        :return: The h parameter.
        :rtype: float
        """
        self._update_canform()
        if self.h_vec is not None:
            return self.h_vec.copy()
        return None

    def get_g(self):
        """
        Get the g parameter.

        :return: The g parameter.
        :rtype: float
        """
        self._update_canform()
        return self.g_val

    def _cov_exists(self):
        """
        Check is the cov matrix has or can be calculated (i.e the precision matrix is not singular).

        :return: The result of the check.
        :rtype: bool
        """
        if self.covform:
            return True
        try:
            np.linalg.inv(self.prec)
        except np.linalg.LinAlgError:
            return False
        return True

    def get_cov(self):
        """
        Get the covariance parameter.

        :return: The cov parameter.
        :rtype: numpy.ndarray
        """
        self._update_covform()
        if self.cov is not None:
            return self.cov.copy()
        return None

    def get_mean(self):
        """
        Get the mean parameter.

        :return: The mean parameter.
        :rtype: numpy.ndarray
        """
        self._update_covform()
        if self.mean is not None:
            return self.mean.copy()
        return None

    def get_log_weight(self):
        """
        Get the log weight parameter.

        :return: The log_weight parameter.
        :rtype: numpy.ndarray
        """
        self._update_covform()
        return self.log_weight

    def normalize(self):
        """
        Normalize the factor.

        :return: The normalized factor.
        :rtype: Gaussian
        """
        gaussian_copy = self.copy()
        gaussian_copy._update_covform()
        gaussian_copy.log_weight = 0.0
        gaussian_copy._destroy_canform()
        return gaussian_copy

    def _get_complex_weight(self):
        """
        Get (possibly complex) weight parameter.

        :return: The weight parameter.
        :rtype: float
        """
        return cmath.exp(self._get_complex_log_weight())

    def _get_complex_log_weight(self):
        """
        Get the log weight even in cases where the determinant of the covariance matrix is negative. In such cases the
        log_weight no longer corresponds to the integral and the log of the weight will have a imaginary component.
        Computing the complex log-weight can however still be useful: it is used, for example, in the (experimental)
        Gaussian mixture division function (_gm_division_m2).

        :return: The potentially complex log weight
        :rtype: complex
        """
        self._update_canform()
        cov = np.linalg.inv(self.prec)
        mean = cov @ self.h_vec
        ut_prec_u = (mean.T @ self.prec @ mean).astype(complex)
        log_weight = self.g_val + 0.5 * ut_prec_u + 0.5 * cmath.log(np.linalg.det(2.0 * np.pi * cov))
        return log_weight

    def _invert(self):
        """
        Invert this Gaussian (1/Gaussian). This is used in the approximate Gaussian mixture division algorithms.

        :return: The inverted Gaussian.
        """
        assert self.canform
        gaussian_copy = self.copy()
        gaussian_copy.prec = (-1.0) * self.prec
        gaussian_copy.h_vec = (-1.0) * self.h_vec
        gaussian_copy.g_val = (-1.0) * self.g_val
        gaussian_copy._destroy_covform()
        return gaussian_copy

    def _add_log_weight(self, log_weight_to_add):
        """
        Add log value to the log weight.

        :param log_weight_to_add: The log value to add to the weight.
        :type log_weight_to_add: float
        """
        if self.canform:
            self.g_val += log_weight_to_add
        if self.covform:
            self.log_weight += log_weight_to_add

    def get_weight(self):
        """
        Get the weight of the distribution - the value of the integral of the (possibly unnormalized) distribution.

        :return: The weight.
        :rtype: float
        """
        return np.exp(self.get_log_weight())

    def _canform_marginalise(self, vars_to_keep, vars_to_integrate_out):
        """
        Calculate the marginal distribution using the canonical parameters.

        :param vars_to_keep: The variables to keep.
        :param vars_to_integrate_out: The variables to integrate out/
        :return: The marginal Gaussian
        """
        xx_indices = [self.var_names.index(var_x) for var_x in vars_to_keep]
        yy_indices = [self.var_names.index(var_y) for var_y in vars_to_integrate_out]
        prec_xx = self.prec[np.ix_(xx_indices, xx_indices)]
        prec_xy = self.prec[np.ix_(xx_indices, yy_indices)]
        prec_yx = prec_xy.T
        prec_yy = self.prec[yy_indices][:, yy_indices]
        inv_prec_yy = np.linalg.inv(prec_yy)
        h_x = self.h_vec[xx_indices]
        h_y = self.h_vec[yy_indices]
        prec_xy_inv_prec_yy = prec_xy @ np.linalg.inv(prec_yy)
        prec = prec_xx - prec_xy_inv_prec_yy @ prec_yx
        h_vec = h_x - prec_xy_inv_prec_yy @ h_y
        log_det_2pi_inv_prec_yy = np.log(np.linalg.det(2.0 * np.pi * inv_prec_yy))
        g_val = self.g_val + 0.5 * (h_y.T @ inv_prec_yy @ h_y + log_det_2pi_inv_prec_yy)
        return Gaussian(prec=prec, h_vec=h_vec, g_val=g_val, var_names=vars_to_keep)

    def marginalize(self, vrs, keep=True):
        """
        Integrate out variables from this Gaussian.

        :param vrs: A subset of variables in the factor's scope.
        :type vrs: str list
        :param keep: Whether to keep or sum out vrs.
        :type keep: bool
        :return: The resulting Gaussian marginal.
        :rtype: Gaussian
        """
        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_integrate_out = list(set(self.var_names) - set(vars_to_keep))
        if self._is_vacuous:
            # TODO: check this (esp log_weight)
            return Gaussian.make_vacuous(var_names=vars_to_keep)

        if self.canform:
            return self._canform_marginalise(vars_to_keep, vars_to_integrate_out)

        assert self.covform
        indices_to_keep = [self._var_names.index(variable) for variable in vars_to_keep]
        marginal_var_names = vars_to_keep.copy()

        marginal_cov = self.cov[np.ix_(indices_to_keep, indices_to_keep)]
        marginal_mean = self.mean[np.ix_(indices_to_keep, [0])]
        return Gaussian(
            cov=marginal_cov, mean=marginal_mean, log_weight=self.log_weight, var_names=marginal_var_names
        )

    def _destroy_canform(self):
        """
        Destroy the canonical form parameters.
        """
        self.prec, self.h_vec, self.h_vec = None, None, None
        self.canform = False

    def _destroy_covform(self):
        """
        Destroy the covariance form parameters.
        """
        self.cov, self.mean, self.log_weight = None, None, None
        self.covform = False

    def _update_canform(self):
        """
        Update the canonical form parameters of the Gaussian.
        """
        if self.canform:
            return
        assert self.covform
        self.prec = _factor_utils.inv_matrix(self.cov)
        self.h_vec = self.prec @ self.mean
        utku = self.mean.T @ self.prec @ self.mean

        det_2pi_cov = np.linalg.det(2.0 * np.pi * self.cov)
        updated_g_val = self.log_weight - 0.5 * utku - 0.5 * _factor_utils.log(det_2pi_cov)

        self.g_val = _factor_utils.make_scalar(updated_g_val)
        self.canform = True

    def _update_covform(self):
        """
        Update the covariance form parameters of the Gaussian.
        """
        if self._is_vacuous:
            raise Exception("cannot update covariance form for vacuous Gaussian.")
        if self.covform:
            return
        assert self.canform
        self.cov = _factor_utils.inv_matrix(self.prec)
        assert not np.isnan(np.sum(self.cov)), "Error: nan values in cov matrix after inversion."
        self.mean = self.cov @ self.h_vec
        utku = self.mean.T @ self.prec @ self.mean
        det_2_pi_cov = np.linalg.det(2.0 * np.pi * self.cov)
        log_weight_ = self.g_val + 0.5 * utku + 0.5 * np.log(det_2_pi_cov)
        self.log_weight = _factor_utils.make_scalar(log_weight_)
        self.covform = True

    def multiply(self, factor):
        """
        Multiply this Gaussian with another factor.

        :param Gaussian factor: the factor to multiply with
        :return: the resulting factor
        """
        # if isinstance(factor, NonLinearGaussian):
        #    return factor.multiply(self)
        return self._absorb_or_cancel(factor, operator.add)

    def divide(self, factor):
        """
        Divide this Gaussian by another factor.

        :param Gaussian factor: the factor to divide by
        :return: the resulting factor
        """
        return self._absorb_or_cancel(factor, operator.sub)

    def _absorb_or_cancel(self, factor, operator_function):
        """
        A general function which can either perform Gaussian multiplication or division (which are very similar).

        :param factor: the gaussian to multiply or divide by
        :param operator_function: the operator to use one the Gaussian canonical parameters
                        ('+' for multiplication and '-' for division)
        :return: the resulting Gaussian
        """
        prec_a = self.get_prec()
        prec_b = factor.get_prec()
        assert len(prec_a.shape) == 2
        assert len(prec_b.shape) == 2
        prec_c, new_vars_0 = _factor_utils.indexed_square_matrix_operation(
            prec_a, prec_b, self._var_names, factor.var_names, operator_function
        )
        g_a = self.get_g()
        g_b = factor.get_g()
        g_c = operator_function(g_a, g_b)

        h_a = self.get_h()
        h_b = factor.get_h()
        h_c, new_vars_1 = _factor_utils.indexed_column_vector_operation(
            h_a, h_b, self._var_names, factor.var_names, operator_function
        )
        assert new_vars_0 == new_vars_1
        return Gaussian(prec=prec_c, h_vec=h_c, g_val=g_c, var_names=new_vars_0)

    def sample(self, num_samples):
        """
        Draw samples from the Gaussian distribution.

        :param num_samples: The number of samples to draw.
        :type num_samples:
        :return: The samples.
        :rtype: int
        """
        std_gaussian_samples = np.random.normal(0, 1, [self.dim, num_samples])
        zero_mean_samples = std_gaussian_samples
        self._update_covform()
        chol_cov = np.linalg.cholesky(self.cov)
        samples = chol_cov @ zero_mean_samples + self.mean
        return samples

    def _get_observation_reduced_canonical_vars(self, observed_indices, unobserved_indices, observed_vec):
        """
        A helper function for reduce.

        :param observed_indices: The observed variable indices
        :type observed_indices: int list
        :param unobserved_indices: The unobserved variable indices
        :type unobserved_indices: int list
        :return: the reduced parameters prec_observed, h_observed, g_observed
        :rtype: numpy.ndarray, numpy.ndarray, float
        """
        prec = self.get_prec()
        h_vec = self.get_h()
        prec_reduced = 0
        h_reduced = 0
        if unobserved_indices:
            prec_xx = prec[np.ix_(unobserved_indices, unobserved_indices)]
            prec_reduced = prec_xx.copy()
            prec_xy = prec[np.ix_(unobserved_indices, observed_indices)]
            h_x = h_vec[np.ix_(unobserved_indices, [0])]
            h_reduced = h_x - prec_xy @ observed_vec

        prec_yy = prec[np.ix_(observed_indices, observed_indices)]
        h_y = h_vec[np.ix_(observed_indices, [0])]
        g_reduced = self.get_g() + h_y.T @ observed_vec - 0.5 * observed_vec.T @ prec_yy @ observed_vec
        return prec_reduced, h_reduced, g_reduced

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this Gaussian and return the resulting factor.

        :param vrs: the names of the observed variable (list)
        :type vrs: str list
        :param values: the values of the observed variables
        :type values: vector-like
        :return: the resulting Gaussian
        :rtype: Gaussian
        """
        observed_vec = _factor_utils.make_column_vector(values)

        assert isinstance(vrs, list)  # just to future-proof interface
        assert set(vrs) <= set(self._var_names), "observed variables must a be subset of the gaussian variables."

        unobserved_vars = list(set(self._var_names) - set(vrs))
        unobserved_vars.sort()  # the above operations seems to return inconsistent orderings
        observed_indices = [self._var_names.index(v) for v in vrs]
        unobserved_indices = [self._var_names.index(v) for v in unobserved_vars]

        prec_reduced, h_reduced, g_reduced = self._get_observation_reduced_canonical_vars(
            observed_indices=observed_indices,
            unobserved_indices=unobserved_indices,
            observed_vec=observed_vec,
        )

        return Gaussian(prec=prec_reduced, h_vec=h_reduced, g_val=g_reduced, var_names=unobserved_vars)

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between this factor and a uniform copy of it.

        :return: The KL divergence.
        :rtype: float
        """
        #  Note: here it does not matter if we take KL(P||Q) or KL(Q||P) the result is either 0.0 (if both are vacuous)
        #        or inf (if one is not).
        if self._is_vacuous:
            return 0.0
        return float("inf")

    def kl_divergence(self, factor, normalize_factor=True):
        """
        Get the KL-divergence D_KL(self || factor) between a normalized version of this factor and another factor.

        :param factor: The other factor
        :type factor: Gaussian
        :param normalize_factor: Whether or not to normalize the other factor before computing the KL-divergence.
        :type normalize_factor: bool
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        if self.dim != factor.dim:
            raise ValueError(
                "cannot calculate KL-divergence between Gaussians of different dimensionalities."
            )
        if self._is_vacuous and factor._is_vacuous:
            return 0.0
        if self._is_vacuous or factor._is_vacuous:
            return np.inf

        if self.equals(factor):
            return 0.0
        # TODO: can we compute the correct ('normalised') KL divergence without explicitly normalizing?
        normalized_self = self.normalize()
        factor_ = factor
        if normalize_factor:
            factor_ = factor.normalize()
        inv_cov_q = factor_.get_prec()
        inv_cov_p = normalized_self.get_prec()
        cov_p = normalized_self.get_cov()

        u_q = factor_.get_mean()
        u_p = normalized_self.get_mean()

        det_inv_cov_q = np.linalg.det(inv_cov_q)
        det_inv_cov_p = np.linalg.det(inv_cov_p)
        assert det_inv_cov_q != 0.0, "Unexpected factor covariance determinant of 0."
        det_term = 0.5 * cmath.log(det_inv_cov_p / det_inv_cov_q)
        trace_term = 0.5 * np.trace(inv_cov_q @ cov_p)
        mahalanobis_term = 0.5 * (u_p - u_q).T @ inv_cov_q @ (u_p - u_q)
        kld = det_term + trace_term + mahalanobis_term - 0.5 * normalized_self.dim
        # TODO: Add warning or error if this is negative and remove abs below
        return np.abs(kld[0][0])

    @property
    def is_vacuous(self):
        """
        Check if a Gaussian distribution contains no information. This is the case when the precision matrix is a zero
         matrix.

        :return: The result of the check.
        :rtype: Bool
        """

        if self._is_vacuous:
            return True
        if self.canform:
            if np.allclose(self.prec, 0.0):
                if not _factor_utils.is_pos_def(self.prec):
                    return True
        return False

    def copy(self):
        """
        Make a copy of this Gaussian.

        :return: The copied Gaussian.
        :rtype: Gaussian
        """

        if self.covform and self.canform:
            assert isinstance(self.g_val, (int, float))
            assert isinstance(self.log_weight, (int, float))
            gaussian_copy = Gaussian(
                cov=self.cov.copy(),
                mean=self.mean.copy(),
                log_weight=self.log_weight,
                var_names=copy.deepcopy(self._var_names),
            )
            gaussian_copy.prec = self.prec.copy()
            gaussian_copy.h_vec = self.h_vec.copy()
            gaussian_copy.g_val = self.g_val
            gaussian_copy.canform = True
            return gaussian_copy

        if self.covform:
            assert isinstance(self.log_weight, (int, float))
            return Gaussian(
                cov=self.cov.copy(),
                mean=self.mean.copy(),
                log_weight=self.log_weight,
                var_names=copy.deepcopy(self._var_names),
            )
        if self.canform:
            assert isinstance(self.g_val, (int, float))
            return Gaussian(
                prec=self.prec.copy(), h_vec=self.h_vec.copy(), g_val=self.g_val,
                var_names=copy.deepcopy(self._var_names)
            )
        raise Exception("Gaussian is neither in canonical form nor in covariance form?")

    def potential(self, x_val):
        """
        Get the value of the Gaussian potential at x_val.

        :param x_val: The vector (or vector-like object) to evaluate the Gaussian at
        :type x_val: numpy.ndarray
        :return: The value of the Gaussian potential at x_val.
        """
        return np.exp(self.log_potential(x_val))

    def log_potential(self, x_val, vrs=None):
        """
        Get the log of the value of the Gaussian potential at x_val.

        :param x_val: the vector (or vector-like object) to evaluate the Gaussian at
        :type x_val: vector-like
        :param vrs: The variables corresponding to the values in x_val.
        :type vrs: str list
        :return: The log of the value of the Gaussian potential at x_val.
        """
        if vrs is not None:
            assert set(vrs) == set(self.var_names)
            if isinstance(x_val, np.ndarray):
                x_val = x_val.ravel()
            x_val_list = list(x_val)
            assert len(x_val_list) == len(self.var_names)
            x_val = [x_val_list[vrs.index(v)] for v in self.var_names]
        x_vec = _factor_utils.make_column_vector(x_val)
        if self.covform:
            log_norm = self.log_weight - 0.5 * np.log(np.linalg.det(2.0 * np.pi * self.cov))
            prec = _factor_utils.inv_matrix(self.cov)
            exponent = -0.5 * (self.mean - x_vec).T @ prec @ (self.mean - x_vec)
            log_potx = log_norm + exponent
        if self.canform:
            log_potx = -0.5 * x_vec.T @ self.prec @ x_vec + x_vec.T @ self.h_vec + self.g_val
        return log_potx[0, 0]

    def _get_cov_repr_str(self):
        """
        Get the string representation for the covariance form parameters.

        :return: The parameter representation string.
        :rtype: str
        """
        self_copy = self.copy()
        self_copy._update_covform()
        np.set_printoptions(linewidth=np.inf)
        repr_str = (
                "cov        = \n"
                + str(self_copy.cov)
                + "\n"
                + "mean       = \n"
                + str(self_copy.mean)
                + "\n"
                + "log_weight = \n"
                + str(self_copy.log_weight)
                + "\n"
        )
        return repr_str

    def _get_can_repr_str(self):
        """
        Get the string representation for the canonical form parameters.

        :return: The parameter representation string.
        :rtype: str
        """
        self_copy = self.copy()
        self_copy._update_canform()
        np.set_printoptions(linewidth=np.inf)
        repr_str = (
                "prec = \n"
                + str(self_copy.prec)
                + "\n"
                + "h = \n"
                + str(self_copy.h_vec)
                + "\n"
                + "g = \n"
                + str(self_copy.g_val)
                + "\n"
                + "is_vacuous: "
                + str(self_copy._is_vacuous)
                + "\n"
        )
        return repr_str

    def __repr__(self):
        """
        Get the string representation of the Gaussian factor.

        :return: The factor representation string.
        :rtype: str
        """
        np.set_printoptions(edgeitems=3)
        np.set_printoptions(precision=4)
        np.core.arrayprint._line_width = 200
        repr_str = "vars = " + str(self.var_names) + "\n"
        if not self._is_vacuous:
            repr_str += self._get_can_repr_str()
        repr_str += self._get_cov_repr_str()
        return repr_str

    def show(self):
        """
        Print the parameters of the Gaussian distribution
        """
        np.set_printoptions(edgeitems=3)
        np.set_printoptions(precision=4)
        np.core.arrayprint._line_width = 200
        print("vars = ", self.var_names)
        if self.covform:
            print(self._get_cov_repr_str())
        if self.canform:
            print(self._get_can_repr_str())

    def show_vis(self, figsize=(10, 8)):
        """
        Visualize the parameters with plots.

        :param figsize: The figure size.
        :type figsize: 2 element int tuple
        """
        # TODO: find a way of making cov matrix square.
        _, [ax_cov, ax_mean] = plt.subplots(nrows=2, figsize=figsize)
        cov_df = pd.DataFrame(data=self.get_cov(), index=self.var_names, columns=self.var_names)
        mean_df = pd.DataFrame(data=self.get_mean(), index=self.var_names, columns=["var_names"])
        cbar_kws = dict(use_gridspec=False, location="top")
        sns.heatmap(cov_df, ax=ax_cov, cbar=True, cbar_kws=cbar_kws, annot=True)
        mean_df.plot.bar(ax=ax_mean, legend=False)
        plt.xticks(rotation=0)

    def _get_limits_for_2d_plot(self):
        """
        Get x and y limits which includes most of the factors mass, by considering
        the mean and variance of each Gaussian component.
        """
        self_copy = self.copy()
        self_copy._update_covform()
        x_lower = self_copy.mean[0, 0] - 3.0 * np.sqrt(self_copy.cov[0, 0])
        x_upper = self_copy.mean[0, 0] + 3.0 * np.sqrt(self_copy.cov[0, 0])
        y_lower = self_copy.mean[1, 0] - 3.0 * np.sqrt(self_copy.cov[1, 1])
        y_upper = self_copy.mean[1, 0] + 3.0 * np.sqrt(self_copy.cov[1, 1])
        return [x_lower, x_upper], [y_lower, y_upper]

    # TODO: reconcile with GaussianMixture _plot_2d
    def _plot_2d(self, log, x_lim=None, y_lim=None):
        """
        Plot a 2d Gaussian potential function

        :param log: if this is True, the log-potential will be plotted
        :param x_lim: the x limits to plot the function over (optional)
        :param y_lim: the y limits to plot the function over (optional)
        """
        xlim_default, ylim_default = self._get_limits_for_2d_plot()
        if x_lim is None:
            x_lim = xlim_default
        if y_lim is None:
            y_lim = ylim_default

        xlabel = self.var_names[0]
        ylabel = self.var_names[1]

        if not log:
            _factor_utils.plot_2d(func=self.potential, xlim=x_lim, ylim=y_lim, xlabel=xlabel, ylabel=ylabel)
        else:
            _factor_utils.plot_2d(func=self.log_potential, xlim=x_lim, ylim=y_lim, xlabel=xlabel, ylabel=ylabel)

    def plot(self, log=False, xlim=None, ylim=None):
        """
        Plot the Gaussian mixture potential function (only for 1d and 2d functions)

        :param log: if this is True, the log-potential will be plotted
        :param xlim: the x limits to plot the function over (optional)
        :type xlim: 2 element float list
        :param ylim: the y limits to plot the function over (optional and only used in 2d case)
        :type ylim: 2 element float list
        """
        # TODO: Replace with gaussian mixture plot.
        if self.dim == 1:
            if xlim is None:
                stddev = np.sqrt(self.get_cov()[0, 0])
                lower_bound = self.get_mean()[0, 0] - 3.0 * stddev
                upper_bound = self.get_mean()[0, 0] + 3.0 * stddev
                xlim = [lower_bound, upper_bound]
            if xlim is not None:
                x_lower = xlim[0]
                x_upper = xlim[1]
            num_points = 200
            x_series = np.linspace(x_lower, x_upper, num_points)
            if log:
                potx = np.array([self.log_potential(xi) for xi in x_series])
            else:
                potx = np.array([self.potential(xi) for xi in x_series])
            plt.plot(x_series, potx)
        elif self.dim == 2:
            self._plot_2d(log=log, x_lim=xlim, y_lim=ylim)
        else:
            raise NotImplementedError("Plotting not implemented for dim!=1.")

    # TODO: Generalise this to more than one dimensional cases.
    def _split_gaussian(self):
        """
        Split this factor into a three component Gaussian Mixture (with all components having different means) where the
        new mixture has the same mean and covariance as the original Gaussian.

        :return: The split Gaussian Mixture.
        :rtype: GaussianMixture
        """
        if self.dim != 1:
            raise NotImplementedError("Gaussian must be one dimensional.")
        weights = [1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0]
        full_cov = self.get_cov()
        full_mean = self.get_mean()

        covs = [weight * full_cov for weight in weights]
        means = [full_mean - np.sqrt(full_cov), full_mean, full_mean + np.sqrt(full_cov)]

        gaussians = []
        for mean, cov, weight in zip(means, covs, weights):
            gaussian = Gaussian(cov=cov, mean=mean, log_weight=np.log(weight), var_names=self.var_names)
            gaussians.append(gaussian)

        return gm.GaussianMixture(gaussians)


class GaussianTemplate(FactorTemplate):
    """
    A class for specifying Gaussian factor templates and creating categorical factors from these templates.
    """

    def __init__(self, gaussian_parameters, var_templates):
        """
        Create a Categorical factor template.

        :param gaussian_parameters: The cononical Gaussian parameters (see example below)
        :type gaussian_parameters: str to np.array/float dict
        :param var_templates: A list of formattable strings.
        :type var_templates: str list

        >>>gaussian_parameters = {'prec': np.array([[1,0],[0,1]]),
        >>>                       'h': np.array([[0],[1]]),
        >>>                       'g': 0}
        """

        super().__init__(var_templates=var_templates)
        self.prec = gaussian_parameters["prec"]
        self.h_vec = gaussian_parameters["h"]
        self.g_val = gaussian_parameters["g"]

    def make_factor(self, format_dict=None, var_names=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.

        :param format_dict: The dictionary to be used to format the var_templates strings.
        :type format_dict: str dict
        :param var_names: The variable names.
        :type var_names: str list
        :return: The instantiated factor.
        :rtype: Gaussian
        """
        if format_dict is not None:
            assert var_names is None
            var_names = [vt.format(**format_dict) for vt in self._var_templates]
        # TODO: remove this and find better solution
        g_val = Gaussian(prec=self.prec.copy(), h_vec=self.h_vec.copy(), g_val=self.g_val, var_names=var_names)
        g_val._is_vacuous = True
        return g_val



"""
A module for factorised factor functionality
"""

from veroku.factors._factor import Factor
from veroku.factors._factor_utils import get_subset_evidence
from veroku._constants import DEFAULT_FACTOR_RTOL, DEFAULT_FACTOR_ATOL

# TODO: This class in very much a work in progress. This needs to be completed. We must think especially carefully about
#   how we multiply factors when we have special factors such as nonlinear Gaussian factors in the factorised factor.
#   Currently this class is mostly focussed on factorised Gaussians.

# pylint: disable=protected-access


class FactorizedFactor(Factor):
    """
    A factorised factor class that allows factors to be used in factorised form - a product of other factors.
    This is useful for improving efficiency between factor multiplication and marginalisation. It also allows the
    for a efficient and intuitive representation of independent factors (especially factors with disjoint scopes).
    """

    # TODO: see how this is used in practice and investigate what the best way is to treat the variable name parameters.
    def __init__(self, factors):
        """
        The initializer.

        :param factors: The list of factors that will be implicitly multiplied.
        """
        assert all([not isinstance(factor, FactorizedFactor) for factor in factors])
        factors_var_name_lists = [factor.var_names for factor in factors]
        factors_var_name = list({var_name for sublist in factors_var_name_lists for var_name in sublist})
        super().__init__(var_names=factors_var_name)
        self.factors = []
        for factor in factors:
            if isinstance(factor, FactorizedFactor):
                self.factors += [f.copy() for f in factor.factors]
            else:
                self.factors.append(factor.copy())

    @property
    def joint_distribution(self):
        """
        The joint distribution.
        """
        if len(self.factors) == 0:
            raise ValueError("FactorisedFactor has no factors.")
        joint_distribution = self.factors[0].copy()
        for factor in self.factors[1:]:
            joint_distribution = joint_distribution.multiply(factor)
        return joint_distribution

    def copy(self):
        """
        Copy this factorised factor.

        :return: the copied factor
        :rtype: FactorizedFactor
        """
        factor_copies = []
        for factor in self.factors:
            factor_copies.append(factor.copy())
        return FactorizedFactor(factor_copies)

    def multiply(self, factor):
        """
        Multiply by another factor.

        :param factor: any type of factor.
        :return: the resulting factor
        :rtype: FactorizedFactor
        """
        # TODO: add preference for factors with conditioning scope? Not sure if practically necessary,
        #  but would be best.
        factorised_factor_copy = self.copy()
        index = factorised_factor_copy._first_factor_with_vars_index(factor.var_names)
        if index is not None:
            factorised_factor_copy.factors[index] = factorised_factor_copy.factors[index].multiply(factor)
        else:
            factorised_factor_copy.factors.append(factor)
        return factorised_factor_copy

    def divide(self, factor):
        """
        Divide out a general factor.

        :param factor: any type of factor.
        :return: the resulting factor
        :rtype: FactorizedFactor
        """
        factorised_factor_copy = self.copy()
        index = factorised_factor_copy._first_factor_with_vars_index(factor.var_names)

        for i, factorised_comp in enumerate(self.factors):
            if factorised_comp.equals(factor):
                if len(self.factors) > 1:
                    del factorised_factor_copy.factors[i]
                    return factorised_factor_copy
                # TODO: add else to make general vacuous factor (wont necessarily be Gaussian)
        if index is not None:
            factorised_factor_copy.factors[index] = factorised_factor_copy.factors[index].divide(factor)
        else:
            raise Exception("Error: Cannot divide factor with disjoint scope.")
        return factorised_factor_copy

    @property
    def is_vacuous(self):
        """
        Check if the factor is vacuous (i.e uniform).

        :return: The result of the check.
        :rtype: Bool
        """
        all_vacuous = all([factor._is_vacuous for factor in self.factors])
        if all_vacuous:
            return True
        none_vacuous = all([not factor._is_vacuous for factor in self.factors])
        if none_vacuous:
            return False
        # TODO: implement this
        raise NotImplementedError()

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between this factor and a uniform copy of it.

        :return: The KL divergence.
        :rtype: float
        """
        # TODO: improve this
        return self.joint_distribution.distance_from_vacuous()

    def kl_divergence(self, factor):
        """
        Get the KL-divergence D_KL(self || factor) between a normalized version of this factor and another factor.

        :param factor: The other factor
        :type factor: Factor
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        # TODO: Make this more efficient.
        return self.joint_distribution.kl_divergence(factor)

    @property
    def num_factors(self):
        """
        Get the number of factors in the factorised factor.

        :return: The number of factors.
        :rtype: int
        """
        return len(self.factors)

    def normalize(self):
        """
        Normalize the factor.

        :return: The normalized factor.
        """
        factor_copy = self.copy()
        factor_copy._merge_dependent_factors()
        for i, factor_i in enumerate(factor_copy.factors):
            factor_copy.factors[i] = factor_i.normalize()
        return factor_copy

    def marginalize(self, vrs, keep=True):
        """
        Marginalize out a subset of the variables in this factor's scope.

        :param list vrs: the variable names
        :param bool keep: whether to keep or sum (or integrate) out these variables.
        :return: the resulting marginal
        :rtype: FactorisedFactor
        """
        # TODO: resolve this: Non-linear Gaussian factors can cause a problem here when there is enough information
        #   that they all should be well defined, but the information resides in only one of them for instance. This
        #   information takes the form of the observed evidence and the conditioning factors.

        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_integrate_out_set = set(self.var_names) - set(vars_to_keep)
        factors_in_keep_scope_indices = self._all_intersecting_factors_indices(vars_to_keep)
        if not factors_in_keep_scope_indices:
            raise Exception("cannot marginalize out all variables from FactorisedFactor.")

        factor_marginals = []
        self_copy = self.copy()
        additional_log_weight = 0.0
        self_copy._merge_factors_with_shared_scope(vars_to_integrate_out_set)
        for factor in self_copy.factors:
            factor_vars_to_integrate_out_set = set(factor.var_names).intersection(vars_to_integrate_out_set)
            if factor_vars_to_integrate_out_set == set(factor.var_names):
                additional_log_weight += factor.get_log_weight()
            else:
                factor_vars_to_integrate_out = list(factor_vars_to_integrate_out_set)
                factor_marginal = factor.marginalize(factor_vars_to_integrate_out, keep=False)
                factor_marginals.append(factor_marginal)
        factor_marginals[0]._add_log_weight(additional_log_weight)
        if len(factor_marginals) == 1:
            return factor_marginals[0]
        marginal_factor = FactorizedFactor(factor_marginals)
        return marginal_factor

    # TODO: make these more efficient by storing joint_distribution
    # TODO: Consider removing or changing this since it will not generalise to cases where the join distribution is not
    #  Gaussian
    def get_cov(self):
        """
        Get the covariance matrix of the joint distribution (if Gaussian).

        :return: The cov parameter.
        :rtype: numpy.ndarray
        """
        return self.joint_distribution.get_cov()

    def get_mean(self):
        """
        Get the mean vector of the joint distribution (if Gaussian).

        :return: The mean parameter.
        :rtype: numpy.ndarray
        """
        return self.joint_distribution.get_mean()

    def _merge_dependent_factors(self):
        """
        Merge all factors (by multiplying) that have overlapping scopes.

        :return: The factor containing the reduced set of factors.
        :rtype: FactorizedFactor
        """
        # TODO: Add tests for this function
        merged_factors = []
        already_merged_factor_indices = []
        for i, factor_i in enumerate(self.factors):
            factor_i_merged = factor_i.copy()
            for j in range(i + 1, len(self.factors)):
                factor_j = self.factors[j]
                if j not in already_merged_factor_indices:
                    if set(factor_i.var_names).intersection(set(factor_j.varnames)) > 0:
                        factor_i_merged = factor_i_merged.multiply(factor_j)
                        already_merged_factor_indices.append(j)
            merged_factors.append(factor_i_merged)
        return FactorizedFactor(merged_factors)

    def _merge_factors_with_shared_scope(self, vrs_set):
        """
        Merge all factors in groups that all contain variables in vrs and have overlap within these subsets. This
        function will typically be used to ensure that all factors containing variables that need to be marginalised out
        are merged together into disjoint (in terms of the integrand variables) factors, so that these larger factors
        can be marginalised independently.

        :param vrs_set: The common scope variables to be considered to determine 'dependent' factors.
        :type vrs_set: str list
        """
        observed_factor_vars_list = []

        for factor in self.factors:
            observed_factor_vars = set(factor.var_names).intersection(vrs_set)
            observed_factor_vars_list.append(observed_factor_vars)

        merged_factors = []
        merged_factors_indices = []
        for i, observed_factor_i_vars in enumerate(observed_factor_vars_list):
            if i not in merged_factors_indices:
                merged_factor = self.factors[i]
                for j in range(i + 1, len(observed_factor_vars_list)):
                    observed_factor_j_vars = observed_factor_vars_list[j]
                    if j not in merged_factors_indices:
                        if len(observed_factor_i_vars.intersection(observed_factor_j_vars)) > 0:
                            merged_factor = merged_factor.multiply(self.factors[j])
                            merged_factors_indices.append(j)
                merged_factors.append(merged_factor)
        self.factors = merged_factors

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this Gaussian and return the resulting factor.

        :param vrs: the names of the observed variable (list)
        :type vrs: str list
        :param values: the values of the observed variables
        :type values: vector-like
        :return: the resulting Gaussian
        :rtype: FactorizedFactor
        """
        all_evidence_dict = dict(zip(vrs, values))
        reduced_factors = []
        additional_log_weight = 0.0
        for factor in self.factors:
            factor_var_names = factor.var_names
            if set(vrs) == set(factor_var_names):
                additional_log_weight += factor.log_potential(x_val=values, vrs=vrs)
            elif len(set(vrs).intersection(set(factor_var_names))) > 0:
                subset_vrs, subset_values = get_subset_evidence(
                    all_evidence_dict=all_evidence_dict, subset_vars=factor.var_names
                )
                reduced_factor = factor.reduce(subset_vrs, subset_values)
                reduced_factors.append(reduced_factor)
            else:
                reduced_factors.append(factor.copy())
        adjusted_weight_factor_0 = reduced_factors[0]
        adjusted_weight_factor_0._add_log_weight(additional_log_weight)
        reduced_factors[0] = adjusted_weight_factor_0
        return FactorizedFactor(reduced_factors)

    def _first_factor_with_vars_index(self, var_names):
        """
        Get the index of the first factor in self_factors with var_names in its scope.

        :param var_names: The variables names to check for.
        :return: The matching factor.
        :rtype: Factor subclass
        """
        # TODO: add check for equals variable names - will be faster for multiplication
        for i, factor in enumerate(self.factors):
            if set(var_names).issubset(set(factor.var_names)):
                return i
        return None

    def _all_intersecting_factors_indices(self, var_names):
        """
        Get the indices of factors in self_factors which has overlap with a set of variable names.

        :param var_names: The variable names
        :type var_names: str list
        :return: The indices
        :rtype: int list
        """
        indices = []
        # TODO: add check for equals variable names - will be faster for multiplication
        for i, factor in enumerate(self.factors):
            if len(set(var_names).intersection(set(factor.var_names))) > 0:
                indices.append(i)
        return indices

    def equals(self, factor, rtol=DEFAULT_FACTOR_RTOL, atol=DEFAULT_FACTOR_ATOL):
        """
        Check if this factor is the same as another factor.

        :param factor: The other factor to compare to.
        :type factor: FactorizedFactor
        :param float rtol: The relative tolerance to use for factor equality check.
        :param float atol: The absolute tolerance to use for factor equality check.
        :return: The result of the comparison.
        :rtype: bool
        """
        if not isinstance(factor, FactorizedFactor):
            # TODO: find a better solution here
            return self.equals(FactorizedFactor([factor]))
        if set(factor.var_names) != set(self.var_names):
            return False
        if len(self.factors) != len(factor.factors):
            return self.joint_distribution.equals(factor.joint_distribution)

        self_comps_with_cpoies_in_factor = []
        for i, self_factor_i in enumerate(self.factors):
            if i not in self_comps_with_cpoies_in_factor:
                for other_factor_j in factor.factors:
                    if self_factor_i.equals(other_factor_j):
                        self_comps_with_cpoies_in_factor.append(i)
        if len(self_comps_with_cpoies_in_factor) == len(self.factors):
            return True
        return False

    def show(self):
        """
        Print this factorised factor.
        """
        for i, factor in self.factors:
            print(f"\n factor {i}/{len(self.factors)}:")
            factor.show()
